#include "light.h"
#include "framebuffer.h"
#include "texture.h"
#include "scene.h"
#include "shader.h"
#include "shadersource.h"
#include "phongmaterial.h"
#include "mesh.h"
#include <resource.h>
#include <glm/ext.hpp>

RESOURCE_PACK(shaders)

const char* Light::flagStr[] = {
    "AMBIENT_LIGHT",
    "DIRECTIONNAL_LIGHT",
    "POINT_LIGHT",
    "SPOT_LIGHT",
    "SHADOWMAP"
};

const glm::mat4 Light::biasMatrix(
        0.5, 0.0, 0.0, 0.0,
        0.0, 0.5, 0.0, 0.0,
        0.0, 0.0, 0.5, 0.0,
        0.5, 0.5, 0.5, 1.0);

Light::Light()
{
    initAmbientLight();
}

void Light::initAmbientLight(glm::vec3 lightColor)
{
    m_type = AMBIENT;
    m_position = glm::vec3(0); // not used
    m_direction = glm::vec3(0, 0, 1); // not used
    m_cutOffAngle = 0; // not used
    m_color = lightColor;
    m_attenuation = 0; // not used
    m_shadowCaster = false; // not used
}

void Light::initDirectionnalLight(glm::vec3 dir, glm::vec3 lightColor)
{
    m_type = DIRECTIONNAL;
    m_position = glm::vec3(0); // not used
    m_direction = dir;
    m_cutOffAngle = 0; // not used
    m_color = lightColor;
    m_attenuation = 0; // not used
    m_shadowCaster = false;
}

void Light::initPointLight(glm::vec3 pos, glm::vec3 lightColor, float att)
{
    m_type = POINT;
    m_position = pos;
    m_direction = glm::vec3(0, 0, 1); // used to orient the shadowmap
    m_cutOffAngle = 0; // not used
    m_color = lightColor;
    m_attenuation = att;
    m_shadowCaster = false;
}

void Light::initSpotLight(glm::vec3 pos, glm::vec3 dir, float spotAngle, glm::vec3 lightColor, float att)
{
    m_type = SPOT;
    m_position = pos;
    m_direction = dir;
    m_cutOffAngle = spotAngle;
    m_color = lightColor;
    m_attenuation = att;
    m_shadowCaster = false;
}

void Light::initShadowMap(int resolution, glm::vec3 dim)
{
    m_shadowMapResolution = resolution;
    Texture* tex;

    switch(m_type)
    {
    case DIRECTIONNAL:
        // 3 orthogonal matrices (cascaded shadowmaps)
        // 3 square textures
        m_viewMatrix.push_back(glm::lookAt(m_position, m_position-m_direction, glm::vec3(0, 1, 0)));
        m_projectionMatrix = glm::ortho(-dim.x/2, dim.x/2, -dim.y/2, dim.y/2, -dim.z/2, dim.z/2);
        tex = new Texture(GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT, resolution, resolution, GL_FLOAT, GL_TEXTURE_2D);
        break;
    case POINT:
        // see http://learnopengl.com/#!Advanced-Lighting/Shadows/Point-Shadows

        // 6 projection matrices
        // 1 cubemap texture
        m_viewMatrix = glm::lookAt(m_position, m_position-glm::vec3(0, 0, 1), glm::vec3(0, 1, 0));
        m_projectionMatrix = glm::perspective(90.f, 1.f, 0.1f, 100.f);
        tex = new Texture(GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT, resolution, resolution, GL_FLOAT, GL_TEXTURE_CUBE_MAP);

        m_viewMatrices.push_back(shadowProj *
                         glm::lookAt(m_position, m_position + glm::vec3(1.0,0.0,0.0), glm::vec3(0.0,-1.0,0.0));
        m_viewMatrices.push_back(shadowProj *
                         glm::lookAt(m_position, m_position + glm::vec3(-1.0,0.0,0.0), glm::vec3(0.0,-1.0,0.0));
        m_viewMatrices.push_back(shadowProj *
                         glm::lookAt(m_position, m_position + glm::vec3(0.0,1.0,0.0), glm::vec3(0.0,0.0,1.0));
        m_viewMatrices.push_back(shadowProj *
                         glm::lookAt(m_position, m_position + glm::vec3(0.0,-1.0,0.0), glm::vec3(0.0,0.0,-1.0));
        m_viewMatrices.push_back(shadowProj *
                         glm::lookAt(m_position, m_position + glm::vec3(0.0,0.0,1.0), glm::vec3(0.0,-1.0,0.0));
        m_viewMatrices.push_back(shadowProj *
                         glm::lookAt(m_position, m_position + glm::vec3(0.0,0.0,-1.0), glm::vec3(0.0,-1.0,0.0));
        break;
    case SPOT:
        // 1 projection matrix
        // 1 square texture
        m_viewMatrix.push_back(glm::lookAt(m_position, m_position+m_direction, glm::vec3(0, 1, 0)));
        m_projectionMatrix = glm::perspective(m_cutOffAngle, 1.f, 0.1f, 100.f);
        tex = new Texture(GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT, resolution, resolution, GL_FLOAT, GL_TEXTURE_2D);
        break;
    case AMBIENT:
        return;
    }

    // shader compilation
    // TODO : do not compile shader for every light
    ShaderSource source;
    Resource::ResourceMap shaderMap;
    Resource::getResourcePack_shaders(shaderMap);
    source.setSource(shaderMap["shaders/shadow.vert.glsl"], ShaderSource::VERTEX);
    if(m_type == POINT)
        source.setSource(shaderMap["shaders/shadow.geom.glsl"], ShaderSource::GEOMETRY);
    source.setSource(shaderMap["shaders/shadow.frag.glsl"], ShaderSource::FRAGMENT);
    m_shaders[0] = source.compile(Mesh::MESH_3D, getFlags());
    m_shaders[1] = source.compile(Mesh::MESH_3D & Mesh::MATERIAL_ALPHA_MASK, getFlags());

    // Depth buffer
    tex->setFiltering(GL_LINEAR);
    tex->setWrap(GL_CLAMP_TO_EDGE);
    tex->setParameter(GL_TEXTURE_COMPARE_MODE, GL_COMPARE_REF_TO_TEXTURE);
    tex->setParameter(GL_TEXTURE_COMPARE_FUNC, GL_LESS);

    // framebuffer
    m_shadowCaster = true;
    m_shadowMap = new FrameBuffer();
    m_shadowMap->addTexture(tex, GL_DEPTH_ATTACHMENT);
    m_shadowMap->initColorAttachments();
}

void Light::generateShadowMap(Scene* scene)
{
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glViewport(0, 0, m_shadowMapResolution, m_shadowMapResolution);
    m_shadowMap->bindFBO();
    glClearDepth(1.0);
    glClear(GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    // glCullFace(GL_FRONT);
    for(SceneIterator<GeometryNode*>* geometryIt = scene->getGeometry();
        geometryIt->isValid(); geometryIt->next())
    {
        GeometryNode* node = geometryIt->getItem();
        if(node->mesh->getFlags() & (1 << Mesh::MESH_SHADOWED))
        {
            // compute matrix attributes
            glm::mat4 lightMVP = getProjectionMatrix() * (getViewMatrix() * node->modelMatrix);
            int hasAlpha = (node->mesh->getFlags() & (1 << Mesh::MATERIAL_ALPHA_MASK)) > 0;
            m_shaders[hasAlpha]->bind();
            m_shaders[hasAlpha]->bindMat4(m_shaders[hasAlpha]->getLocation("MVP"), lightMVP);
            node->mesh->draw(m_shaders[hasAlpha], false, hasAlpha, false);
        }
    }
    // glCullFace(GL_BACK);
}

void Light::bindShadowMap(Shader *shader)
{
    glm::mat4 lightMVP = Light::biasMatrix * m_projectionMatrix * light->getViewMatrix() * node->modelMatrix;
    shader->bindMat4(shader->getLocation("lightMVP"), lightMVP);
}

Texture* Light::getShadowMapTexture()
{
    return m_shadowMap->getTexture(0);
}

void Light::setPosition(glm::vec3 new_pos)
{
    m_position = new_pos;
    m_viewMatrix = glm::lookAt(m_position, m_position+m_direction, glm::vec3(0, 1, 0));
}

unsigned int Light::getFlags()
{
    unsigned int flags = 0;
    switch(m_type)
    {
    case AMBIENT :
        flags = 1 << AMBIENT_FLAG;
        break;
    case DIRECTIONNAL :
        flags = 1 << DIRECTIONNAL_FLAG;
        break;
    case POINT :
        flags = 1 << POINT_FLAG;
        break;
    case SPOT :
        flags = 1 << SPOT_FLAG;
        break;
    }
    if(m_shadowCaster)
        flags |= 1 << SHADOWMAP_FLAG;
    return flags;
}
