#ifndef LIGHT_H
#define LIGHT_H

#include <glm/vec3.hpp>
#include <glm/mat4x4.hpp>
#include <string>
#include <vector>

class FrameBuffer;
class Shader;
class Scene;
class Texture;

class Light
{
public:
    enum LightType {
        AMBIENT,
        DIRECTIONNAL,
        POINT,
        SPOT
    };

    enum LightFlags {
        AMBIENT_FLAG,
        DIRECTIONNAL_FLAG,
        POINT_FLAG,
        SPOT_FLAG,
        SHADOWMAP_FLAG,
        NB_FLAGS
    };
    static const char* flagStr[];
    static const glm::mat4 biasMatrix;

    virtual LightType getType() = 0;

    glm::vec3 getColor() { return m_color; }

    virtual bool isShadowCaster() { return false; }

    virtual void bindAttributes(Shader *shader) = 0;

    virtual unsigned int getFlags() = 0;

protected:
    glm::vec3 m_color;
};

class AmbientLight : public Light
{
public:
    AmbientLight(glm::vec3 lightColor = glm::vec3(0.1f)) : m_color(lightColor) {}
    virtual LightType getType() { return AMBIENT; }
    virtual void bindAttributes(Shader *shader)
        { shader->bindVec3(shader->getLocation("lightColor"), m_color); }
    virtual unsigned int getFlags() { return 1 << AMBIENT_FLAG; }
};

class ShadowableLight : public Light
{
public:
    // enables shadowmapping on this light
    virtual void initShadowMap(int resolution, glm::vec3 dim = glm::vec3(1)) = 0;

    // disables shadowmapping on this light
    virtual void destroyShadowMap() = 0;

    // returns true is shadowmapping is enabled on this light
    virtual bool isShadowCaster() { return m_shadowCaster; }

    // updates the shadowmap if shadowmapping is enabled
    virtual void updateShadowMap(Scene* scene) = 0;

protected:
    bool m_shadowCaster;
};

class Light
{
public:

    Light();
    void initAmbientLight(glm::vec3 lightColor = glm::vec3(0.1f));
    void initDirectionnalLight(glm::vec3 dir = glm::vec3(0, -1, 0), glm::vec3 lightColor = glm::vec3(1));
    void initPointLight(glm::vec3 pos = glm::vec3(0), glm::vec3 lightColor = glm::vec3(1), float att = 1);
    void initSpotLight(glm::vec3 pos = glm::vec3(0), glm::vec3 dir = glm::vec3(1, 0, 0), float spotAngle = 360, glm::vec3 lightColor = glm::vec3(1), float att = 1);


    glm::vec3 getDir() { return m_direction; }
    glm::vec3 getPos() { return m_position; }
    glm::vec3 getColor() { return m_color; }
    float getAttenuation() { return m_attenuation; }
    float getCutoffAngle() { return m_cutOffAngle; }

    bool isShadowCaster() {return m_shadowCaster;}
    void initShadowMap(int resolution, glm::vec3 dim = glm::vec3(1));
    void generateShadowMap(Scene* scene);
    void bindShadowMap(Shader *shader);
    Texture* getShadowMapTexture();

    void setAttenuation(float a) { m_attenuation = a; }
    void setCutoffAngle(float c) { m_cutOffAngle = c; }
    void setPosition(glm::vec3 new_pos);
    void setColor(glm::vec3 new_color) {m_color = new_color;}

    /**
     * @brief getFlags returns the flags that defines the specificities of the light
     */
    unsigned int getFlags();
private:
    // standard attributes
    LightType m_type;

    glm::vec3 m_position;
    glm::vec3 m_direction;
    float m_cutOffAngle;
    float m_attenuation;
    glm::vec3 m_color;

    // shadowmap attributes
    bool m_shadowCaster;
    int m_shadowMapResolution;
    FrameBuffer* m_shadowMap;
    Shader* m_shaders[2];
    std::vector<glm::mat4> m_viewMatrices;
    glm::mat4 m_projectionMatrix;
};

#endif // LIGHT_H
